from __future__ import print_function

import sys
import json
import csv
import collections

def openfile(outfile):
    if sys.version_info[0] < 3:
        return open(outfile, 'wb')
    else:
        return open(outfile, 'w')

class Report2Table ():
    def __init__(self):
        pass

    def convert(self, reportjsonfile, outfile):
        print('converting json report to tabular csv')
        json_rep = None
        with open(reportjsonfile) as repfile:
            json_rep = json.load(repfile, object_pairs_hook=collections.OrderedDict)
        
        with openfile(outfile) as csvfile:
            cwriter = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL)
            cwriter.writerow(['Table'])


            apps = list(json_rep['apps'].keys())
            traces = list(json_rep['summaries'].keys())

            summary_metrics = [] if len(traces)<1 else list(json_rep['summaries'][traces[0]].keys())
            var_metrics = [] if len(apps)<1 else list(json_rep['apps'][apps[0]].keys())

            ind_stats = []
            if len(apps)>0 and len(var_metrics)>0 and len(traces)>0:
                ind_stats = list(json_rep['apps'][apps[0]][var_metrics[0]][traces[0]].keys())
            ind_st = [''] * (len(ind_stats) * len(traces))
            ind_st[0::len(traces)] = ind_stats

            compare_stats = []
            compare_st = []
            compare_rels = list(json_rep['apps'][apps[0]][var_metrics[0]]["comparison"].keys())
            if len(compare_rels) > 0 :
                compare_stats = list(json_rep['apps'][apps[0]][var_metrics[0]]["comparison"][compare_rels[0]].keys())
                compare_st = [''] * (len(compare_stats) * len(compare_rels))
                compare_st[0::len(compare_rels)] = compare_stats

            # print('apps', apps)
            # print('var_metrics', var_metrics)
            # print('traces', traces)
            # print('ind_stats', ind_stats)
            # print('ind_st', ind_st)
            # print('compare_stats', ind_stats)
            # print('compare_st', ind_st)

            cwriter.writerow([])
            cwriter.writerow(['Single Valued Metrics'])
            cwriter.writerow([''] + summary_metrics)
            for tr in traces:
                vals = [tr]
                for metric_v in summary_metrics:
                    vals.append(json_rep['summaries'][tr][metric_v])
                cwriter.writerow(vals)

            cwriter.writerow([])
            cwriter.writerow(['Time Varying Metrics'])
            for app in apps:
                cwriter.writerow([])
                cwriter.writerow([app])

                cwriter.writerow([''] + ind_st + compare_st)
                cwriter.writerow([''] + traces * len(ind_stats) + compare_rels * len(compare_stats))
                
                for metric_v in var_metrics:
                    row = [metric_v]

                    for st in ind_stats:
                        for tr in traces:
                            if tr in json_rep['apps'][app][metric_v] and st in json_rep['apps'][app][metric_v][tr]:
                                row.append(json_rep['apps'][app][metric_v][tr][st])
                            else:
                                row.append('')

                    for cst in compare_rels:
                        for cr in compare_stats:
                            if metric_v in json_rep['apps'][app] and cr in json_rep['apps'][app][metric_v]['comparison'][cst]:
                                row.append(json_rep['apps'][app][metric_v]['comparison'][cst][cr])
                            else:
                                row.append('')

                    cwriter.writerow(row)

            cwriter.writerow([])

if __name__ == '__main__':
    Report2Table().convert(sys.argv[1], sys.argv[2])
