# Copyright (c) 2017  University of Houston

import os
import sys
import math
import time
import json
import logging
import argparse
import shutil
import warnings
from collections import OrderedDict, defaultdict
import numpy as np
import pandas as pd
import scipy
import scipy.stats
from sklearn.metrics import mean_squared_error
import pywt
from pcap2csv.pcap2csv import Pcap2csv
from hurst import compute_Hc
from RegscorePy.aic import aic
from RegscorePy.bic import bic
from pcap2csv.pcap2csvutils import Pcap2csvutils
from .const import conn_pkts_reqd_metrics, pkt_graph_metrics, pdu_graph_metrics
from .const import per_pkt_metrics, metric_units, time_resolution

try:
    warnings.simplefilter(action="ignore", category=pd.core.common.SettingWithCopyWarning)
except AttributeError:
    warnings.simplefilter(action="ignore", category=pd.errors.SettingWithCopyWarning)


logger = logging.getLogger(__name__)

def fmt(val):
    if isinstance(val,float):
        return float('%.5f' % (val))
    return val

def remove_zeros(x, y):
    #logger.debug 'len x & y', len(x), len(y)
    d = {'x':x,'y':y}
    dff = pd.DataFrame(d)
    mdf = dff[dff['y'] != 0]
    xx = mdf.x.tolist()
    yy = mdf.y.tolist()
    #logger.debug 'len xx & yy', len(xx), len(yy)
    return xx, yy

def remove_nan(x, y):
    df = pd.DataFrame({'x':x,'y':y})
    df = df.dropna()
    return df.x.tolist(), df.y.tolist()

def myconverter(obj):
    if isinstance(obj, dict) or isinstance(obj, OrderedDict):
        for key,val in obj.items():
            obj[key]=myconverter(val)
        return obj
    elif isinstance(obj, list):
        for val in obj:
            val=myconverter(val)
        return obj
    elif isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.float64):
        return float(obj)
    else:
        return obj

class TmMetricsUtilsUsess():
    def __init__(self):
        pass

    #utils
    def t_start(self, pkt_list):
        return pkt_list.iloc[0].time
        
    def t_end(self, pkt_list):
        return pkt_list.iloc[-1].time

    def getStartStopTimes(self, pkt_lists):
        #print('len(pkt_lists)', len(pkt_lists))
        start = self.t_start(pkt_lists[0])
        stop = self.t_end(pkt_lists[0])
        
        for pkt_list in pkt_lists:
            start = min(start, self.t_start(pkt_list))
            stop = max(start, self.t_end(pkt_list))
        return (start, stop) 

    def getXpsListUsess(self, df_t_val):
        df_t_val['time_bac'] = df_t_val.time
        usess_xys = []
        usess_grps = df_t_val.groupby('user_sess_idx')
        for i, upkts in usess_grps:
            if len(upkts.index) == 0:
                continue
            upkts.time = upkts.time - upkts.iloc[0].time
            x, y = self.getXpsList(upkts)
            usess_xys.append((x,y))
        df_t_val.time = df_t_val.time_bac
        return usess_xys

    def getXpsList(self, df_t_val):
        '''fast accurate use vectorization'''

        if len(df_t_val) <= 1:
            return [[], []]
        #print(df_t_val.dtypes['val'], df_t_val['val'])
        df_t_val = df_t_val[np.isfinite(df_t_val['val'])]
        #dp = 1
        #print ('df_t_val', type(df_t_val))
        #print (df_t_val)

        step = time_resolution * 1.0
        dec = 1.0/step

        if len(df_t_val) <= 1:
            return [[], []]
        #dec, step = 10**dp, 1./(10**dp)
        start, stop = self.getStartStopTimes([df_t_val])
        #logger.debug 'start:', start, 'stop:', stop
        #print 'usess_idx', df_t_val.user_sess_idx

        ret_df = pd.DataFrame([])
        xps_times = np.trunc(np.arange(start, stop + step + step, step)*dec)/dec
        xps_times = np.unique(xps_times)
        xps_labels = xps_times[:-1].tolist()
        ret_df['time'] = xps_times[:-1]
        ret_df['val'] = 0

        #print 'xps_times', xps_times

        #print

        #print 'xps_labels', np.array(xps_labels)
        #print xps_times
        
        cutter = pd.cut(df_t_val.time, xps_times.tolist(), labels=xps_labels).astype(float)

        #print(type(cutter))
        #print 'cutter'
        #print cutter

        #logger.debug 'len cutter', len(cutter.index), 'cutter-firstisna', pd.isna(cutter.iloc[0])
        if len(cutter.index)>0 and pd.isna(cutter.iloc[0]): 
            #print '---true'
            cutter.iloc[0]=0.0
            
        #print 'cutter'
        #print cutter
        #print

        grp = df_t_val.groupby([cutter])

        #print 'grp', 
        #for gid, pkts in grp:
        #    print gid
        #    print pkts
        #print 

        grp_starts = grp.first().time
        #print 'grp_starts'
        #print grp_starts
        #print

        intvls = []
        for i in range(len(grp_starts)):#-1):
            q = math.ceil(grp_starts.iloc[i]* dec)/dec
            if i < len(grp_starts)-1 and grp_starts.iloc[i+1] < q + step :
                f = grp_starts.iloc[i+1]
                m = grp_starts.iloc[i+1] - grp_starts.iloc[i]
                #print 
            else:
                f = q + step
                m = q - grp_starts.iloc[i]
            m = step #can remove this.
            intvls.append(m)

            #print grp_starts.iloc[i], f


        #intvls.append(math.ceil(grp_starts.iloc[-1]) - grp_starts.iloc[-1])

        #print 'intvls', 
        #print intvls
        #print 

        grp_intvals = pd.Series(intvls)
        grp_vals = grp[['val']].sum().reset_index()
        grp_vals.val = grp_vals.val /grp_intvals
        ret_df.set_index('time', inplace=True)
        ret_df.update(grp_vals.set_index('time'))
        ret_df.reset_index(inplace=True)


        #print ret_df.time.tolist()
        #print ret_df.val.tolist()

        #print (ret_df.head())        
        x = [float(i) for i in ret_df.time.tolist()]
        y = [float(i) for i in ret_df.val.tolist()]

        return x, y
        
    def getXpsList0(self, df_t_val):
        '''much slower also accurate'''
        start_time = time.time()
        #return a list of packets per seconds
        #df.val = df.val.fillna(0)
        df_t_val = df_t_val[np.isfinite(df_t_val['val'])]   
        #print( 'ppppppppppp\n', df_t_val.val)
        if len(df_t_val) <= 1:
            return [[], []]
        #dp = 1 #decimal places for interval
        step = time_resolution * 1.0
        dec = 1.0/step
        #dec, step = 10**dp, 1./(10**dp)
        start, stop = self.getStartStopTimes([df_t_val])
        logger.debug('before getting xps times --- time elapsed = ' + str(time.time() - start_time))
        xps_times = (np.trunc(np.arange(start, stop + step,
                                        step)*dec)/dec).tolist()
        logger.debug('before getting xps groups 1 --- time elapsed = ' + str(time.time() - start_time))
        groups = df_t_val.groupby(pd.cut(df_t_val.time, xps_times, duplicates='drop'))
        logger.debug('before getting xps groups 2 --- time elapsed = ' + str(time.time() - start_time))
        groupsx = df_t_val.groupby(pd.cut(df_t_val.time, xps_times, duplicates='drop'), as_index=False)
        
        logger.debug('before getting xps starts --- time elapsed = ' + str(time.time() - start_time))
        starts = groupsx.first().time
        replace = pd.DataFrame(xps_times[:-1])
        replace = replace[replace.columns[0]]
        logger.debug('after xps replace --- time elapsed = ' + str(time.time() - start_time))
        starts[starts.isnull()] = replace
        widths = groups.first().time.diff().shift(-1)
        logger.debug('after xps widths 1 --- time elapsed = ' + str(time.time() - start_time))
        last_int = df_t_val.iloc[-1].time - groups.first().iloc[-1].time
        widths.iloc[-1] = last_int
        logger.debug('after xps widths 2 --- time elapsed = ' + str(time.time() - start_time))
        xps_list = groups.val.sum()
        #print(type(starts), starts)
        #print('-------------\n\n')

        logger.debug('after inner xps_list --- time elapsed = ' + str(time.time() - start_time))
        widths = widths.replace(0, step)
        widths = widths.fillna(step)
        logger.debug('after xps widths3 --- time elapsed = ' + str(time.time() - start_time))
        xps_list = xps_list*1/widths
        xps_list = xps_list.fillna(0)

        logger.debug('before xps_list return --- time elapsed = ' + str(time.time() - start_time))
        return starts.tolist()[:-1], xps_list.tolist()[:-1]

    def get_xy_vals(self, pkt_pdus_df, metric):
        if metric in ['select_metric']:
            logger.debug('not implemented %s', metric)
            return [], []
        else:
            metric_function = getattr(self, 'get_xy_vals_' + metric)
            x,y = metric_function(pkt_pdus_df)
            x,y = remove_nan(x,y)
            return x, y

    def get_xy_vals_usess(self, pkt_pdus_df, metric):
        if metric in [ 'select_metric']:
            logger.debug('%s not implemented', metric)
            return []
        else:
            metric_function = getattr(self, 'get_xy_vals_usess_' + metric)
            usess_xys = metric_function(pkt_pdus_df)
            n_usess_xys = []
            for x,y in usess_xys:
                xx,yy = remove_nan(x,y)
                n_usess_xys.append((xx, yy))
            return n_usess_xys

    #packet metrics single val
    def get_summary_pkts(self, pkts_df):
        logger.debug('obtaining filtered pkt summary')
        dur = 0
        if len(pkts_df.index) > 0:
            dur = pkts_df.time.iloc[-1] - pkts_df.time.iloc[0]
        
        ret = OrderedDict()
        ret['total_packets'] = len(pkts_df)
        ret['duration'] = dur
        ret['l2_total_payload_bytes'] = pkts_df.l2_d_size.sum()
        ret['l2_total_vlan_packets'] = len(pkts_df[pd.notnull(pkts_df['l2_vlan_idx'])])
        ret['l2_number_of_vlans'] = len(pkts_df[pd.notnull(pkts_df['l2_vlan_idx'])].l2_vlan_idx.unique())
        ret['l3_total_payload_bytes'] = pkts_df.l3_d_size.sum()
        ret['l3_fragmented_packets'] = len( pkts_df[((pkts_df['l3_mf'] ==1) | (pkts_df['l3_frag_offset'] > 0))] )
        ret['l4_total_payload_bytes'] = pkts_df.l4_d_size.sum()
        if 'l4_ack' in pkts_df: #replace with if any tcp pkt in df
            ret['l4_total_retransmitted_pkts'] = len( pkts_df[pkts_df['l4_is_retransmission'] == 1])
            ret['l4_total_outoforder_pkts'] = len( pkts_df[pkts_df['l4_is_out_of_order'] == 1])
            ret['l4_total_acks'] = len( pkts_df[pkts_df['l4_ack'] == 1])
            ret['l4_total_push'] = len( pkts_df[pkts_df['l4_psh'] == 1])
        ret['total_conns'] = len(pkts_df.l4_sess_idx.unique())
        return ret

    def get_summary_pkts_usess(self, pkts_df):
        logger.debug('obtaining filtered pkt summary per usess')
        tret = OrderedDict()
        for m in  ['total_packets', 'duration', 'l2_total_payload_bytes', 
                    'l2_total_vlan_packets', 'l2_number_of_vlans', 
                    'l3_total_payload_bytes', 'l3_fragmented_packets', 
                    'l4_total_payload_bytes', 'l4_total_retransmitted_pkts', 
                    'l4_total_outoforder_pkts', 'l4_is_out_of_order',
                    'l4_total_acks', 'l4_total_push', 'total_conns']:
            tret[m] = []

        for i, upkts in pdus_df.groupby('user_sess_idx'):
            if len(upkts.index) == 0:
                continue
            ret = self.get_summary_pkts(upkts)
            for k, v in ret.items():
                tret[k].append(v)
        return tret

    #pdu metrics single val
    def get_summary_pdus(self, pdus_df):
        #logger.debug('obtaining filtered pkt summary')
        dur = 0
        if len(pdus_df.index) > 0:
            dur = pdus_df.time.iloc[-1] - pdus_df.time.iloc[0]
        
        ret = OrderedDict()
        ret['total_PDUs'] = len(pdus_df)
        ret['duration'] = dur
        ret['total_conns'] = len(pdus_df.l4_sess_idx.unique().tolist())
        return ret

    def get_summary_pdus_usess(self, pdus_df):
        logger.debug('obtaining filtered pdu summary per usess')
        tret = OrderedDict()
        for m in  ['total_PDUs', 'duration', 'total_conns']:
            tret[m] = []
        for i, updus in pdus_df.groupby('user_sess_idx'):
            if len(updus.index) == 0:
                continue
            ret = self.get_summary_pdus(updus)
            for k, v in ret.items():
                tret[k].append(v)
        return tret

    #packet metrics series
    def get_xy_vals_packet_rate(self, pkts_df):
        logger.debug('calculating xy vals -  packet rate')
        #df = pkts_df[['pkt_idx', 'time']]
        pkts_df['val'] = 1
        x,y = self.getXpsList(pkts_df)
        return x, y

    def get_xy_vals_usess_packet_rate(self, pkts_df):
        logger.debug('calculating xy vals - packet rates - per usess')
        pkts_df['val'] = 1
        usess_xys = self.getXpsListUsess(pkts_df)
        #print 'usess_xys', usess_xys
        #print
        return usess_xys

    def get_xy_vals_packet_rate_no_zeros(self, pkts_df):
        logger.debug('calculating xy vals - packet rate no zeros')
        pkts_df['val'] = 1
        x,y = self.getXpsList(pkts_df)
        #print y
        xx, yy = remove_zeros(x, y)
        print
        #print yy
        return xx, yy

    def get_xy_vals_usess_packet_rate_no_zeros(self, pkts_df):
        logger.debug('calculating xy vals - packet rates no zero - per usess ')
        pkts_df['val'] = 1
        usess_xys = self.getXpsListUsess(pkts_df)
        new_usess_xys = []
        for x,y in usess_xys:
            xx, yy = remove_zeros(x, y)
            new_usess_xys.append((xx, yy))
        #print new_usess_xys
        return new_usess_xys

    def get_xy_vals_inter_packet_time(self, pkts_df):
        logger.debug('calculating xy vals - app inter pkt times')
        xy = []
        rconn_grps = pkts_df.groupby('l4_sess_idx')
        for i, grp in rconn_grps:
            #print i, len(grp.index)
            prevtime = -1
            for idx, pdu in grp.iterrows():
                if (prevtime != -1): 
                    xy.append((pdu['time'], pdu['time'] - prevtime))
                    prev_is_from_client = False
                prevtime = pdu['time']
        xy = sorted(xy, key=lambda v:v[0])
        #print xy
        x, y = [[ i for i, j in xy ], 
                [ j for i, j in xy ]] 

        x = [float(i) for i in x]
        y = [float(i) for i in y]

        return x, y

    def get_xy_vals_usess_inter_packet_time(self, pkts_df):
        logger.debug('calculating xy vals - app inter packet times - per usess')
        usess_xys = []
        for m, upkts in pkts_df.groupby('user_sess_idx'):
            if len(upkts.index) == 0:
                continue
            upkts['time_bac'] = upkts.time - upkts.iloc[0].time
            xy = []
            #reqs = upkts.loc[upkts['l4_from_client']==True]
            rconn_grps = upkts.groupby('l4_sess_idx')
            for i, grp in rconn_grps:
                #print i, len(grp.index)
                prevtime = -1
                for idx, pdu in grp.iterrows():
                    if (prevtime != -1): 
                        xy.append((pdu['time_bac'], pdu['time_bac'] - prevtime))
                        prev_is_from_client = False
                    prevtime = pdu['time_bac']
            xy = sorted(xy, key=lambda v:v[0])
            #print xy
            x, y = [[ i for i, j in xy ], 
                    [ j for i, j in xy ]] 
            usess_xys.append((x, y))
        return usess_xys

    def get_xy_vals_throughput(self, pkts_df):
        '''
        return a list of bits per seconds on wire
        ignoring vlan header
        '''
        #TODO consider presense of VLAN headers...
        logger.debug('calculating xy vals - total throughput')
        #df = pkts_df[['pkt_idx', 'time']]
        pkts_df['val'] = (pkts_df.l2_d_size + 14) * 8  #14 is MACs+type, 38 is preamble+MACs+type+FCS
        pkts_df['val'].fillna(0) 
        x,y = self.getXpsList(pkts_df)
        x = [float(i) for i in x]
        y = [float(i) for i in y]
        return x, y

    def get_xy_vals_usess_throughput(self, pkts_df):
        logger.debug('calculating xy vals - total throughput - per usess')
        pkts_df['val'] = (pkts_df.l2_d_size + 14) * 8
        pkts_df['val'].fillna(0)
        usess_xys = self.getXpsListUsess(pkts_df)
        #logger.debugx, y
        return usess_xys

    def get_xy_vals_throughput_no_zeros(self, pkts_df):
        logger.debug('calculating xy vals - total throughput no zeros')
        pkts_df['val'] = (pkts_df.l2_d_size + 14) * 8
        pkts_df['val'].fillna(0)
        x,y = self.getXpsList(pkts_df)
        xx,yy = remove_zeros(x, y)
        return xx, yy

    def get_xy_vals_usess_throughput_no_zeros(self, pkts_df):
        logger.debug('calculating xy vals - total throughput no zeros - per usess')
        pkts_df['val'] = (pkts_df.l2_d_size + 14) * 8
        pkts_df['val'].fillna(0)
        usess_xys = self.getXpsListUsess(pkts_df)
        new_usess_xys = []
        for x,y in usess_xys:
            xx, yy = remove_zeros(x, y)
            new_usess_xys.append((xx, yy))
        return new_usess_xys

    def get_xy_vals_packet_size(self, pkts_df):        
        '''
        return a list of packet sizes
        ignoring vlan header
        '''
        #TODO consider presense of VLAN headers...
        logger.debug('calculating xy vals -  packetsize')
        if len(pkts_df)==0:
            return [], []
        m = pkts_df.l2_d_size.fillna(0) + 14 #14 is MACs+type
        x,y = pkts_df.time.tolist(), m.tolist()

        x = [float(i) for i in x]
        y = [float(i) for i in y]

        return x, y

    def get_xy_vals_usess_packet_size(self, pkts_df):
        logger.debug('calculating xy vals - packet size - per usess')
        usess_xys = []
        for i, upkts in pkts_df.groupby('user_sess_idx'):
            if len(upkts.index) == 0:
                continue
            tt = upkts.time - upkts.iloc[0].time
            m = pkts_df.l2_d_size.fillna(0) + 14  #14 is MACs+type
            x,y = tt.tolist(), m.tolist()
            usess_xys.append((x,y))
        return usess_xys

    def get_xy_vals_l2_payload_throughput(self, pkts_df):
        '''
        return a list of bits per seconds on wire
        ignoring vlan header
        '''
        #TODO consider presense of VLAN headers...
        logger.debug('calculating xy vals - l2_payload throughput')
        #df = pkts_df[['pkt_idx', 'time']]
        pkts_df['val'] = pkts_df.l2_d_size * 8  #14 is MACs+type, 38 is preamble+MACs+type+FCS
        pkts_df['val'].fillna(0) 
        x,y = self.getXpsList(pkts_df)
        x = [float(i) for i in x]
        y = [float(i) for i in y]
        return x, y

    def get_xy_vals_usess_l2_payload_throughput(self, pkts_df):
        logger.debug('calculating xy vals - l2_payload throughput - per usess')
        pkts_df['val'] = pkts_df.l2_d_size * 8
        pkts_df['val'].fillna(0)
        usess_xys = self.getXpsListUsess(pkts_df)
        #logger.debugx, y
        return usess_xys

    def get_xy_vals_l2_payload_throughput_no_zeros(self, pkts_df):
        logger.debug('calculating xy vals - l2_payload throughput no zeros')
        pkts_df['val'] = pkts_df.l2_d_size * 8
        pkts_df['val'].fillna(0)
        x,y = self.getXpsList(pkts_df)
        xx,yy = remove_zeros(x, y)
        return xx, yy

    def get_xy_vals_usess_l2_payload_throughput_no_zeros(self, pkts_df):
        logger.debug('calculating xy vals - l2_payload throughput no zeros - per usess')
        pkts_df['val'] = pkts_df.l2_d_size * 8
        pkts_df['val'].fillna(0)
        usess_xys = self.getXpsListUsess(pkts_df)
        new_usess_xys = []
        for x,y in usess_xys:
            xx, yy = remove_zeros(x, y)
            new_usess_xys.append((xx, yy))
        return new_usess_xys
    
    def get_xy_vals_l2_payload_size(self, pkts_df):        
        '''
        return a list of l2 payload sizes
        '''
        #TODO consider presense of VLAN headers...
        logger.debug('calculating xy vals -  l2 payload sizes')
        if len(pkts_df)==0:
            return [], []
        m = pkts_df.l2_d_size.fillna(0)
        x,y = pkts_df.time.tolist(), m.tolist()

        x = [float(i) for i in x]
        y = [float(i) for i in y]

        return x, y

    def get_xy_vals_usess_l2_payload_size(self, pkts_df):
        logger.debug('calculating xy vals - l2 payload size - per usess')
        usess_xys = []
        for i, upkts in pkts_df.groupby('user_sess_idx'):
            if len(upkts.index) == 0:
                continue
            tt = upkts.time - upkts.iloc[0].time
            m = upkts.l2_d_size.fillna(0)
            x,y = tt.tolist(), m.tolist()
            
            usess_xys.append((x,y))
        return usess_xys

    def get_xy_vals_l3_payload_throughput(self, pkts_df):
        '''
        return a list of l3 bits per seconds on wire
        '''
        #TODO consider presense of VLAN headers...
        logger.debug('calculating xy vals - l3_payload throughput')
        #df = pkts_df[['pkt_idx', 'time']]
        pkts_df['val'] = pkts_df.l3_d_size * 8 
        pkts_df['val'].fillna(0) 
        x,y = self.getXpsList(pkts_df)
        x = [float(i) for i in x]
        y = [float(i) for i in y]
        return x, y

    def get_xy_vals_usess_l3_payload_throughput(self, pkts_df):
        logger.debug('calculating xy vals - l3_payload throughput - per usess')
        pkts_df['val'] = pkts_df.l3_d_size * 8
        pkts_df['val'].fillna(0)
        usess_xys = self.getXpsListUsess(pkts_df)
        #logger.debugx, y
        return usess_xys

    def get_xy_vals_l3_payload_throughput_no_zeros(self, pkts_df):
        logger.debug('calculating xy vals - l3_payload throughput no zeros')
        pkts_df['val'] = pkts_df.l3_d_size * 8
        pkts_df['val'].fillna(0)
        x,y = self.getXpsList(pkts_df)
        xx,yy = remove_zeros(x, y)
        return xx, yy

    def get_xy_vals_usess_l3_payload_throughput_no_zeros(self, pkts_df):
        logger.debug('calculating xy vals - l3_payload throughput no zeros - per usess')
        pkts_df['val'] = pkts_df.l3_d_size * 8
        pkts_df['val'].fillna(0)
        usess_xys = self.getXpsListUsess(pkts_df)
        new_usess_xys = []
        for x,y in usess_xys:
            xx, yy = remove_zeros(x, y)
            new_usess_xys.append((xx, yy))
        return new_usess_xys

    def get_xy_vals_l3_payload_size(self, pkts_df):        
        '''
        return a list of l3 payload sizes
        '''
        #TODO consider presense of VLAN headers...
        logger.debug('calculating xy vals - l3 payload size')
        if len(pkts_df)==0:
            return [], []
        m = pkts_df.l3_d_size.fillna(0)
        x,y = pkts_df.time.tolist(), m.tolist()

        x = [float(i) for i in x]
        y = [float(i) for i in y]

        return x, y

    def get_xy_vals_usess_l3_payload_size(self, pkts_df):
        logger.debug('calculating xy vals - l3 payload size - per usess')
        usess_xys = []
        for i, upkts in pkts_df.groupby('user_sess_idx'):
            if len(upkts.index) == 0:
                continue
            tt = upkts.time - upkts.iloc[0].time
            m = upkts.l3_d_size.fillna(0)
            x,y = tt.tolist(), m.tolist()
            
            usess_xys.append((x,y))
        return usess_xys

    def get_xy_vals_l4_payload_throughput(self, pkts_df):
        '''
        return a list of l4 bits per seconds on wire
        '''
        #TODO consider presense of VLAN headers...
        logger.debug('calculating xy vals - l4_payload throughput')
        #df = pkts_df[['pkt_idx', 'time']]
        pkts_df['val'] = pkts_df.l4_d_size * 8 
        pkts_df['val'].fillna(0) 
        x,y = self.getXpsList(pkts_df)
        x = [float(i) for i in x]
        y = [float(i) for i in y]
        return x, y

    def get_xy_vals_usess_l4_payload_throughput(self, pkts_df):
        logger.debug('calculating xy vals - l4_payload throughput - per usess')
        pkts_df['val'] = pkts_df.l4_d_size * 8
        pkts_df['val'].fillna(0)
        usess_xys = self.getXpsListUsess(pkts_df)
        #logger.debugx, y
        return usess_xys

    def get_xy_vals_l4_payload_throughput_no_zeros(self, pkts_df):
        logger.debug('calculating xy vals - l4_payload throughput no zeros')
        pkts_df['val'] = pkts_df.l4_d_size * 8
        pkts_df['val'].fillna(0)
        x,y = self.getXpsList(pkts_df)
        xx,yy = remove_zeros(x, y)
        return xx, yy

    def get_xy_vals_usess_l4_payload_throughput_no_zeros(self, pkts_df):
        logger.debug('calculating xy vals - l4_payload throughput no zeros - per usess')
        pkts_df['val'] = pkts_df.l4_d_size * 8
        pkts_df['val'].fillna(0)
        usess_xys = self.getXpsListUsess(pkts_df)
        new_usess_xys = []
        for x,y in usess_xys:
            xx, yy = remove_zeros(x, y)
            new_usess_xys.append((xx, yy))
        return new_usess_xys

    def get_xy_vals_l4_payload_size(self, pkts_df):        
        '''
        return a list of l4 payload sizes
        '''
        #TODO consider presense of VLAN headers...
        logger.debug('calculating xy vals - l4 payload size')
        if len(pkts_df)==0:
            return [], []
        m = pkts_df.l4_d_size.fillna(0)
        x,y = pkts_df.time.tolist(), m.tolist()

        x = [float(i) for i in x]
        y = [float(i) for i in y]

        return x, y

    def get_xy_vals_usess_l4_payload_size(self, pkts_df):
        logger.debug('calculating xy vals - l4 payload size - per usess')
        usess_xys = []
        for i, upkts in pkts_df.groupby('user_sess_idx'):
            if len(upkts.index) == 0:
                continue
            tt = upkts.time - upkts.iloc[0].time
            m = upkts.l4_d_size.fillna(0)
            x,y = tt.tolist(), m.tolist()
            
            usess_xys.append((x,y))
        return usess_xys

    def get_xy_vals_l4_tcp_retransmitted_rate(self, pkts_df):
        '''get retransmitted pkts per second in TCP pkts'''
        logger.debug('calculating xy vals -  layer 4 retransmitted packet rates')
        x,y = [],[]
        if 'l4_is_retransmission' in pkts_df:
            pkts_df['val'] = 0
            pkts_df.l4_is_retransmission.fillna(False, inplace=True)
            pkts_df.loc[pkts_df.l4_is_retransmission, 'val'] = 1
            x,y = self.getXpsList(pkts_df)
        return x, y
 
    def get_xy_vals_usess_l4_tcp_retransmitted_rate(self, pkts_df):
        logger.debug('calculating xy vals - layer 4 retransmitted packet rates - per usess')
        if 'l4_is_retransmission' in pkts_df:
            pkts_df['val'] = 0
            pkts_df.l4_is_retransmission.fillna(False, inplace=True)
            pkts_df.loc[pkts_df.l4_is_retransmission, 'val'] = 1
            usess_xys = self.getXpsListUsess(pkts_df)
            return usess_xys
        else:
            return []

    def get_xy_vals_l4_outoforder_rate(self, pkts_df):
        '''get retransmitted pkts per second in TCP pkts'''
        logger.debug('calculating xy vals -  layer 4 out of order packet rates')
        x,y = [],[]
        if 'l4_is_out_of_order' in pkts_df:
            pkts_df['val'] = 0
            pkts_df.l4_is_out_of_order.fillna(False, inplace=True)
            #print 'pkts_df.l4_is_out_of_order', pkts_df.l4_is_out_of_order
            #print
            pkts_df.loc[pkts_df.l4_is_out_of_order, 'val'] = 1
            #print 'pkts_df.val', pkts_df.val
            #print
            x,y = self.getXpsList(pkts_df)
        return x, y

    def get_xy_vals_usess_l4_outoforder_rate(self, pkts_df):
        logger.debug('calculating xy vals - l4_outoforder_rates - per usess')
        if 'l4_is_out_of_order' in pkts_df:
            pkts_df['val'] = 0
            pkts_df.l4_is_out_of_order.fillna(False, inplace=True)
            #print 'pkts_df.l4_is_out_of_order', pkts_df.l4_is_out_of_order
            #print
            pkts_df.loc[pkts_df.l4_is_out_of_order, 'val'] = 1
            usess_xys = self.getXpsListUsess(pkts_df)
            return usess_xys
        else:
            return []

    def get_xy_vals_l4_window_size(self, pkts_df):
        logger.debug('calculating xy vals -  layer 4 packets window sizes')
        if len(pkts_df) == 0:
            return {'x':[], 'y':[]}
        x,y = [],[]
        if 'l4_win_size' in pkts_df:
            winsizetimelist = pkts_df[['time', 'l4_win_size']].dropna()
            times = winsizetimelist['time'].tolist()
            winsizes = winsizetimelist['l4_win_size'].tolist()
            x,y = times, winsizes
        return x, y

    def get_xy_vals_usess_l4_window_size(self, pkts_df):
        logger.debug('calculating xy vals - l4 payload size - per usess')
        usess_xys = []
        if 'l4_win_size' in pkts_df:
            for i, upkts in pkts_df.groupby('user_sess_idx'):
                winsizetimelist = pkts_df[['time', 'l4_win_size']].dropna()
                times = winsizetimelist['time'].tolist()
                winsizes = winsizetimelist['l4_win_size'].tolist()
                x,y = times, winsizes
                usess_xys.append((x,y))
        return usess_xys

    def get_xy_vals_l4_unacked_inflight_data(self, pkts_df):
        pass

    #PDU metrics series
    def get_xy_vals_app_pdu_rate(self, pdus_df):
        logger.debug('calculating xy vals - app pdu rates')
        pdus_df['val'] = 1
        x,y = self.getXpsList(pdus_df)
        return x, y

    def get_xy_vals_usess_app_pdu_rate(self, pdus_df):
        logger.debug('calculating xy vals - app pdu rates - per usess')
        pdus_df['val'] = 1
        usess_xys = self.getXpsListUsess(pdus_df)
        #print 'usess_xys', usess_xys
        #print
        return usess_xys

    def get_xy_vals_app_pdu_rate_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app pdu rate no zeros')
        pdus_df['val'] = 1
        x,y = self.getXpsList(pdus_df)
        #print y
        xx, yy = remove_zeros(x, y)
        #print
        #print yy
        return xx, yy

    def get_xy_vals_usess_app_pdu_rate_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app pdu rates no zero - per usess ')
        pdus_df['val'] = 1
        usess_xys = self.getXpsListUsess(pdus_df)
        new_usess_xys = []
        for x,y in usess_xys:
            xx, yy = remove_zeros(x, y)
            new_usess_xys.append((xx, yy))
        #print new_usess_xys
        return new_usess_xys
    
    def get_xy_vals_app_request_rate(self, pdus_df):
        logger.debug('calculating xy vals - app request rates _')
        dfx = pdus_df[pdus_df['l4_from_client']==True]
        return self.get_xy_vals_app_pdu_rate(dfx)

    def get_xy_vals_usess_app_request_rate(self, pdus_df):
        logger.debug('calculating xy vals - app request rates - per usess _')
        dfx = pdus_df[pdus_df['l4_from_client']==True]
        return self.get_xy_vals_usess_app_pdu_rate(dfx)

    def get_xy_vals_app_request_rate_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app request rate no zeros _')
        dfx = pdus_df[pdus_df['l4_from_client']==True]
        return self.get_xy_vals_app_pdu_rate_no_zeros(dfx)

    def get_xy_vals_usess_app_request_rate_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app request rates no zero - per usess _')
        dfx = pdus_df[pdus_df['l4_from_client']==True]
        return self.get_xy_vals_usess_app_pdu_rate_no_zeros(dfx)
        
    def get_xy_vals_app_response_rate(self, pdus_df):
        logger.debug('calculating xy vals - app response rates _')
        dfx = pdus_df[pdus_df['l4_from_client']!=True]
        return self.get_xy_vals_app_pdu_rate(dfx)

    def get_xy_vals_usess_app_response_rate(self, pdus_df):
        logger.debug('calculating xy vals - app response rates - per usess _')
        dfx = pdus_df[pdus_df['l4_from_client']!=True]
        return self.get_xy_vals_usess_app_pdu_rate(dfx)

    def get_xy_vals_app_response_rate_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app response rate no zeros _')
        dfx = pdus_df[pdus_df['l4_from_client']!=True]
        return self.get_xy_vals_app_pdu_rate_no_zeros(dfx)
        
    def get_xy_vals_usess_app_response_rate_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app response rates no zero - per usess _')
        dfx = pdus_df[pdus_df['l4_from_client']!=True]
        return self.get_xy_vals_usess_app_pdu_rate_no_zeros(dfx)

    def get_xy_vals_app_pdu_size(self, pdus_df):
        logger.debug('calculating xy vals - app pdu size')
        x,y = pdus_df.time.tolist(), pdus_df.d_size.fillna(0).tolist()
           
        x = [float(i) for i in x]
        y = [float(i) for i in y]

        return x, y

    def get_xy_vals_usess_app_pdu_size(self, pdus_df):
        logger.debug('calculating xy vals - app pdu size - per usess')
        usess_xys = []
        for i, upkts in pdus_df.groupby('user_sess_idx'):
            if len(upkts.index) == 0:
                continue
            tt = upkts.time - upkts.iloc[0].time
            x,y = tt.tolist(), upkts.d_size.fillna(0).tolist()
            usess_xys.append((x,y))
        return usess_xys

    def get_xy_vals_app_inter_pdu_time(self, pdus_df):
        logger.debug('calculating xy vals - app inter PDU times in conns')
        xy = []
        #reqs = pdus_df.loc[pdus_df['l4_from_client']==True]
        rconn_grps = pdus_df.groupby('l4_sess_idx')
        for i, grp in rconn_grps:
            #print i, len(grp.index)
            prevtime = -1
            for idx, pdu in grp.iterrows():
                if (prevtime != -1): 
                    xy.append((pdu['time'], pdu['time'] - prevtime))
                    prev_is_from_client = False
                prevtime = pdu['time']
        xy = sorted(xy, key=lambda v:v[0])
        #print xy
        x, y = [[ i for i, j in xy ], 
                [ j for i, j in xy ]] 

        x = [float(i) for i in x]
        y = [float(i) for i in y]

        return x, y

    def get_xy_vals_usess_app_inter_pdu_time(self, pdus_df):
        logger.debug('calculating xy vals - app inter PDU time - per usess')
        usess_xys = []
        for m, upkts in pdus_df.groupby('user_sess_idx'):
            if len(upkts.index) == 0:
                continue
            upkts['time_bac'] = upkts.time - upkts.iloc[0].time
            xy = []
            #reqs = upkts.loc[upkts['l4_from_client']==True]
            rconn_grps = upkts.groupby('l4_sess_idx')
            for i, grp in rconn_grps:
                #print i, len(grp.index)
                prevtime = -1
                for idx, pdu in grp.iterrows():
                    if (prevtime != -1): 
                        xy.append((pdu['time_bac'], pdu['time_bac'] - prevtime))
                        prev_is_from_client = False
                    prevtime = pdu['time_bac']
            xy = sorted(xy, key=lambda v:v[0])
            #print xy
            x, y = [[ i for i, j in xy ], 
                    [ j for i, j in xy ]] 
            usess_xys.append((x, y))
        return usess_xys

    def get_xy_vals_app_inter_request_time(self, pdus_df):
        logger.debug('calculating xy vals - app inter PDU times in conns')
        dfx = pdus_df[pdus_df['l4_from_client']==True]
        return self.get_xy_vals_app_inter_pdu_time(dfx)

    def get_xy_vals_usess_app_inter_request_time(self, pdus_df):
        logger.debug('calculating xy vals - app inter PDU time - per usess')
        dfx = pdus_df[pdus_df['l4_from_client']==True]
        return self.get_xy_vals_usess_app_inter_pdu_time(dfx)

    def get_xy_vals_app_inter_response_time(self, pdus_df):
        logger.debug('calculating xy vals - app inter PDU times in conns')
        dfx = pdus_df[pdus_df['l4_from_client']!=True]
        return self.get_xy_vals_app_inter_pdu_time(dfx)

    def get_xy_vals_usess_app_inter_response_time(self, pdus_df):
        logger.debug('calculating xy vals - app inter PDU time - per usess')
        dfx = pdus_df[pdus_df['l4_from_client']!=True]
        return self.get_xy_vals_usess_app_inter_pdu_time(dfx)

    def get_xy_vals_app_pdu_throughput(self, pdus_df):
        logger.debug('calculating xy vals - app pdu throughput')

        pdus_df['val'] = pdus_df['d_size'] * 8
        pdus_df['val'].fillna(0)
        x,y = self.getXpsList(pdus_df)
        #logger.debugx, y
                   
        x = [float(i) for i in x]
        y = [float(i) for i in y]

        return x, y

    def get_xy_vals_usess_app_pdu_throughput(self, pdus_df):
        logger.debug('calculating xy vals - app pdu throughput - per usess')
        pdus_df['val'] = pdus_df['d_size'] * 8
        pdus_df['val'].fillna(0)
        usess_xys = self.getXpsListUsess(pdus_df)
        #logger.debugx, y
        return usess_xys

    def get_xy_vals_app_pdu_throughput_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app pdu throughput no zeros')
        pdus_df['val'] = pdus_df['d_size'] * 8
        pdus_df['val'].fillna(0)
        x,y = self.getXpsList(pdus_df)
        xx,yy = remove_zeros(x, y)
        return xx, yy

    def get_xy_vals_usess_app_pdu_throughput_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app pdu throughput no zeros - per usess')
        pdus_df['val'] = pdus_df['d_size'] * 8
        pdus_df['val'].fillna(0)
        usess_xys = self.getXpsListUsess(pdus_df)
        new_usess_xys = []
        for x,y in usess_xys:
            xx, yy = remove_zeros(x, y)
            new_usess_xys.append((xx, yy))
        return new_usess_xys

    

    def get_xy_vals_app_request_throughput(self, pdus_df):
        logger.debug('calculating xy vals - app request throughput')
        dfx = pdus_df[pdus_df['l4_from_client']==True]
        return self.get_xy_vals_app_pdu_throughput(dfx)

    def get_xy_vals_usess_app_request_throughput(self, pdus_df):
        logger.debug('calculating xy vals - app request throughput - per usess')
        dfx = pdus_df[pdus_df['l4_from_client']==True]
        return self.get_xy_vals_usess_app_pdu_throughput(dfx)

    def get_xy_vals_app_request_throughput_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app request throughput no zeros')
        dfx = pdus_df[pdus_df['l4_from_client']==True]
        return self.get_xy_vals_app_pdu_throughput_no_zeros(dfx)
        
    def get_xy_vals_usess_app_request_throughput_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app request throughput no zeros - per usess')
        dfx = pdus_df[pdus_df['l4_from_client']==True]
        return self.get_xy_vals_usess_app_pdu_throughput_no_zeros(dfx)





    def get_xy_vals_app_response_throughput(self, pdus_df):
        logger.debug('calculating xy vals - app response throughput')
        dfx = pdus_df[pdus_df['l4_from_client']!=True]
        return self.get_xy_vals_app_pdu_throughput(dfx)

    def get_xy_vals_usess_app_response_throughput(self, pdus_df):
        logger.debug('calculating xy vals - app response throughput - per usess')
        dfx = pdus_df[pdus_df['l4_from_client']!=True]
        return self.get_xy_vals_usess_app_pdu_throughput(dfx)

    def get_xy_vals_app_response_throughput_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app response throughput no zeros')
        dfx = pdus_df[pdus_df['l4_from_client']!=True]
        return self.get_xy_vals_app_pdu_throughput_no_zeros(dfx)
        
    def get_xy_vals_usess_app_response_throughput_no_zeros(self, pdus_df):
        logger.debug('calculating xy vals - app response throughput no zeros - per usess')
        dfx = pdus_df[pdus_df['l4_from_client']!=True]
        return self.get_xy_vals_usess_app_pdu_throughput_no_zeros(dfx)





    def get_xy_vals_app_srv_resp_time(self, pdus_df):
        logger.debug('calculating xy vals - app server response times')
        xy = []
        conn_grps = pdus_df.groupby('l4_sess_idx')
        for i, grp in conn_grps:
            prevreqtime = -1
            prev_is_from_client = False
            for idx, pdu in grp.iterrows():
                if (prevreqtime != -1 and pdu['l4_from_client'] == False
                  and prev_is_from_client): 
                    xy.append((pdu['time'], pdu['time'] - prevreqtime))
                    prev_is_from_client = False
                else:
                    prevreqtime = pdu['time']
                    prev_is_from_client = True
        xy = sorted(xy, key=lambda v:v[0])
        x, y = [[ i for i, j in xy ], 
                [ j for i, j in xy ]] 
        return x, y

    def get_xy_vals_usess_app_srv_resp_time(self, pdus_df):
        logger.debug('calculating xy vals - app server response times - per usess')
        usess_xys = []
        for m, upkts in pdus_df.groupby('user_sess_idx'):
            if len(upkts.index) == 0:
                continue
            upkts['time_bac'] = upkts.time - upkts.iloc[0].time
            xy = []
            conn_grps = upkts.groupby('l4_sess_idx')
            for i, grp in conn_grps:
                prevreqtime = -1
                prev_is_from_client = False
                for idx, pdu in grp.iterrows():
                    if (prevreqtime != -1 and pdu['l4_from_client'] == False
                    and prev_is_from_client): 
                        xy.append((pdu['time_bac'], pdu['time_bac'] - prevreqtime))
                        prev_is_from_client = False
                    else:
                        prevreqtime = pdu['time_bac']
                        prev_is_from_client = True
            xy = sorted(xy, key=lambda v:v[0])
            x, y = [[ i for i, j in xy ], 
                    [ j for i, j in xy ]] 
            usess_xys.append((x, y))
        return usess_xys

    #Conn metric series
    def get_xy_vals_conn_inter_arrival_time(self, pkt_pdus_df):
        logger.debug('calculating xy vals - conn inter arrival times')
        dfi = pkt_pdus_df[pkt_pdus_df['l4_sess_idx'].notna()]
        conn_firsts = dfi.groupby('l4_sess_idx').first()
        times = conn_firsts.reset_index().time.sort_values()
        x = times[1:].tolist()
        y = times.diff()[1:].tolist()
        return x, y

    def get_xy_vals_usess_conn_inter_arrival_time(self, pkt_pdus_df):
        logger.debug('calculating xy vals - conn inter arrival times - per usess')
        usess_xys = []
        for m, upkts in pkt_pdus_df.groupby('user_sess_idx'):
            if len(upkts.index) == 0:
                continue

            dfi = upkts[upkts['l4_sess_idx'].notna()]
            conn_firsts = dfi.groupby('l4_sess_idx').first()
            times = conn_firsts.reset_index().time.sort_values()
            x = times[1:].tolist()
            y = times.diff()[1:].tolist()
            usess_xys.append((x, y))
        return usess_xys

    def get_xy_vals_conn_arrival_rate_no_zeros(self, pkt_pdus_df):
        logger.debug('calculating xy vals - conn arrival rates')
        dfi = pkt_pdus_df[pkt_pdus_df['l4_sess_idx'].notna()]
        conn_firsts = dfi.groupby('l4_sess_idx').first()
        conn_firsts['val'] = 1

        x,y = self.getXpsList(conn_firsts)
        #print y
        xx, yy = remove_zeros(x, y)
        #print
        #print yy
        return xx, yy

    def get_xy_vals_usess_conn_arrival_rate_no_zeros(self, pkt_pdus_df):
        logger.debug('calculating xy vals - conn arrival rates - per usess')

        dfi = pkt_pdus_df[pkt_pdus_df['l4_sess_idx'].notna()]
        conn_firsts = dfi.groupby('l4_sess_idx').first()

        conn_firsts['val'] = 1
        usess_xys = self.getXpsListUsess(conn_firsts)
        new_usess_xys = []
        m = 0
        for x,y in usess_xys:
            xx, yy = remove_zeros(x, y)
            new_usess_xys.append((xx, yy))
            m += 1

        return new_usess_xys

    #filters
    def get_filtered_df(self, pkt_pdus_df, filter_str=None, conv='all', endpt='all', app='all'):
        #TODO: URGENT** implement real filters
        logger.debug('filter_str is : %s',filter_str)
        filter_dict = json.loads(filter_str) if filter_str else {}
        logger.debug('filtering packets based on filter string')
        #TODO implement other filters 
        cond = True
        if filter_dict != {}:
            logger.debug('in 1')
            if 'time' in filter_dict:
                cond = cond & (pkt_pdus_df['time'] >= filter_dict['time'][0])
                cond = cond & (pkt_pdus_df['time'] <= filter_dict['time'][1])
            if 'src_mac' in filter_dict:
                cond = cond & (pkt_pdus_df['l2_src_mac'] == filter_dict['l2_src_mac'])
            if 'dst_mac' in filter_dict:
                cond = cond & (pkt_pdus_df['l2_dst_mac'] == filter_dict['l2_dst_mac'])
            if 'src_ip' in filter_dict:
                cond = cond & (pkt_pdus_df['l3_src_ip'] == filter_dict['l3_src_ip'])
            if 'dst_ip' in filter_dict:
                cond = cond & (pkt_pdus_df['l3_dst_ip'] == filter_dict['l3_dst_ip'])
            if 'src_port' in filter_dict:
                cond = cond & (pkt_pdus_df['l4_src_port']==int(filter_dict['l4_src_port']))
            if 'dst_port' in filter_dict:
                cond = cond & (pkt_pdus_df['l4_dst_port']==int(filter_dict['l4_dst_port']))

            filtredpkts = pkt_pdus_df[cond]
        else: 
            filtredpkts =  pkt_pdus_df

        #print (str(app))
        #print (str(endpt))
        #print (str(conv))
        if str(app) != 'all':
            #logger.debug('in 2')
            filtredpkts = filtredpkts.loc[filtredpkts['app_idx']==app]
            #logger.debug ('filtered app %s', str(app))

        if str(endpt) != 'all':
            #logger.debug('in 3')
            filtredpkts = self.filter_by_endpoint(endpt, filtredpkts)
            #logger.debug ('filtered endpt %s', str(endpt))

        if str(conv) != 'all':
            #logger.debug('in 4')
            filtredpkts = self.filter_by_conversation(conv, filtredpkts)
            #logger.debug ('filtered conv %s', str(conv))

        return filtredpkts

    def filter_by_endpoint(self, endpt, filteredpkts):
        if len(filteredpkts.index)==0:
            return filteredpkts
        layer, vals = endpt.split(': ', 1)
        if layer == 'layer4-rx':
            ip, port = vals.split(':', 1)  
            port = int(float(port))
            ff = filteredpkts.loc[(filteredpkts['l3_dst_ip']==ip) & (filteredpkts['l4_dst_port']==port)]
        elif layer == 'layer4-tx':
            ip, port = vals.split(':', 1)
            port = int(float(port))
            ff = filteredpkts.loc[(filteredpkts['l3_src_ip']==ip) & (filteredpkts['l4_src_port']==port)]
        elif layer == 'layer3-rx':
            ip = vals
            ff = filteredpkts.loc[filteredpkts['l3_dst_ip']==ip]
        elif layer == 'layer3-tx':
            ip = vals
            ff = filteredpkts.loc[filteredpkts['l3_src_ip']==ip]
        elif layer == 'layer2-rx':
            mac = vals
            ff = filteredpkts.loc[filteredpkts['l2_dst_mac']==mac]
        elif layer == 'layer2-tx':
            mac = vals
            ff = filteredpkts.loc[filteredpkts['l2_src_mac']==mac]
        else:
            ff = filteredpkts
            #throw error?
        return ff

    def filter_by_conversation(self, conv, filteredpkts):
        if len(filteredpkts.index)==0:
            return filteredpkts
        layer, vals = conv.split(': ', 1)
        if layer == 'layer4':
            endpt1, endpt2 = vals.split(' <-> ',1)
            ip1, port1 = endpt1.split(':',1)
            ip2, port2 = endpt2.split(':',1)
            port1, port2 = int(float(port1)), int(float(port2))
            ff = filteredpkts.loc[((filteredpkts['l3_src_ip'] == ip1)&(filteredpkts['l4_src_port'] == port1)& \
                                   (filteredpkts['l3_dst_ip'] == ip2)&(filteredpkts['l4_dst_port'] == port2)) | \
                                  ((filteredpkts['l3_src_ip'] == ip2)&(filteredpkts['l4_src_port'] == port2)& \
                                   (filteredpkts['l3_dst_ip'] == ip1)&(filteredpkts['l4_dst_port'] == port1))] 
        elif layer == 'layer3':
            ip1, ip2 = vals.split(' <-> ',1)
            ff = filteredpkts.loc[((filteredpkts['l3_src_ip'] == ip1)&(filteredpkts['l3_dst_ip'] == ip2)) |
                                  ((filteredpkts['l3_src_ip'] == ip2)&(filteredpkts['l3_dst_ip'] == ip1))]
        elif layer == 'layer2':
            mac1, mac2 = vals.split(' <-> ',1)
            ff = filteredpkts.loc[((filteredpkts['l2_src_mac'] == mac1)&(filteredpkts['l2_dst_mac'] == mac2)) |
                                  ((filteredpkts['l2_src_mac'] == mac2)&(filteredpkts['l2_dst_mac'] == mac1))]
        else:
            ff = filteredpkts
        return ff

    #stats
    def get_stats(self, x, y, metric):
        logger.debug('obtaining metric statistics')
        my_list = y
        if len(my_list) == 0:
            return {}    
        try:
            ar = np.array(my_list).astype(np.float)
        except AttributeError:
            ar = np.array(my_list).astype(np.float64)

        ar = ar[~np.isnan(ar)]
        my_mean = myconverter(np.mean(ar))
        my_std = myconverter(np.std(ar))
        my_median = myconverter(np.median(ar))
        my_mode = myconverter(scipy.stats.mode(ar))
        my_min = myconverter(min(ar)) if len(ar) > 0 else None
        my_max = myconverter(max(ar)) if len(ar) > 0 else None
        H = None #'~ (len less than 100)'
        #print (metric, ar)
        if len(ar) > 99:
            try:
                H, c, dd = compute_Hc(ar, kind='change', simplified=True)
                H = myconverter(H)
            except FloatingPointError:
                H = None
        ret =  OrderedDict()
        ret ['dataset_len'] = len(y)
        ret['mean'] = fmt(my_mean )
        ret['median'] = fmt(my_median)
        #ret['mode'] = my_mode
        ret['std'] = fmt(my_std)
        ret['max'] = fmt(my_max)
        ret['min'] = fmt(my_min)
        ret['range'] = fmt(myconverter(np.ptp(ar)))
        ret['index of dispersion'] =  fmt(my_std * my_std / my_mean) if my_mean>0 else None
        ret['inter-quartile-range'] = fmt(myconverter(scipy.stats.iqr(ar)))
        ret['skewness'] =  fmt(myconverter(scipy.stats.skew(ar)))
        ret['kurtosis'] =  fmt(myconverter(scipy.stats.kurtosis(ar)))
        ret['Hurst exponent'] = fmt(H)
        #logger.debug('*** len mylist = %d', len(ar))
        #print(ar[:20])
        ret['25th_percentile'] = fmt(myconverter(np.percentile(ar, 25)))
        ret['50th_percentile'] = fmt(myconverter(np.percentile(ar, 50)))
        ret['75th_percentile'] = fmt(myconverter(np.percentile(ar, 75)))

        return ret

    def get_stats_usess(self, usess_xys, metric):
        logger.debug('obtaining metric statistics by usess')

        #if metric == 'app_pdu_rate_no_zeros':
        #    print 'usess_xys',usess_xys

        ret = OrderedDict()
        for m in ['mean','median','std', 'min', '25th_percentile',
                  '50th_percentile','75th_percentile','max',
                  'dataset_len', 'inter-quartile-range',
                  #'index of dispersion',
                  #'skewness','kurtosis', 
                  'Hurst exponent',
                  'self-KS-2-samp-s', 'self-KS-2-samp-pvals']:
            ret[m] = []

        for usess_xy in usess_xys:
            x, y = usess_xy
            my_list = y
            if len(my_list) == 0:
                return {}    
            try:
                ar = np.array(my_list).astype(np.float)
            except AttributeError:
                ar = np.array(my_list).astype(np.float64)
            my_mean = myconverter(np.mean(ar))
            my_std = myconverter(np.std(ar))
            my_median = myconverter(np.median(ar))
            #my_mode = myconverter(scipy.stats.mode(ar))
            my_min = myconverter(min(ar)) if len(ar) > 0 else None
            my_max = myconverter(max(ar)) if len(ar) > 0 else None
            H = None #'~ (len less than 100)'
            if len(ar) > 99:
                try:
                    H, c, dd = compute_Hc(ar, kind='change', simplified=True)
                    H = myconverter(H)
                except FloatingPointError:
                    H = None
            
            ret ['dataset_len'].append(len(y))
            ret['mean'].append(fmt(my_mean ))
            ret['median'].append(fmt(my_median))
            ret['std'].append(fmt(my_std))
            ret['max'].append(fmt(my_max))
            ret['min'].append(fmt(my_min))
            #ret['index of dispersion'].append( fmt(myconverter(my_std * my_std / my_mean)) if my_mean>0 else None)
            ret['inter-quartile-range'].append(fmt(myconverter(scipy.stats.iqr(ar))))
            #ret['skewness'].append( fmt(myconverter(scipy.stats.skew(ar))))
            #ret['kurtosis'].append( fmt(myconverter(scipy.stats.kurtosis(ar))))
            ret['Hurst exponent'].append(fmt(H))
            ret['25th_percentile'].append(fmt(myconverter(np.percentile(my_list, 25))))
            ret['50th_percentile'].append(fmt(myconverter(np.percentile(my_list, 50))))
            ret['75th_percentile'].append(fmt(myconverter(np.percentile(my_list, 75))))

        #print 'len usess', len(usess_xys)
        for i in range(len(usess_xys)):
            for j in range(i+1, len(usess_xys)):
                #print 'i,j', i, j
                try:
                    #print(0)
                    x1, y1 = usess_xys[i]
                    #print(1)
                    x2, y2 = usess_xys[j]
                    #print(2)
                    ijks = scipy.stats.ks_2samp(y1, y2)
                    #print(3)
                    ret['self-KS-2-samp-s'].append(fmt(myconverter(ijks.statistic)))
                    #print(4)
                    ret['self-KS-2-samp-pvals'].append(myconverter(fmt(ijks.pvalue)))
                    #print(5)
                except:
                    #print('error')
                    pass

        #print 'ret', ret['self-KS-2-samp-s']
        #print 'ret', ret['self-KS-2-samp-pvals']
        return ret

    def get_comp_metrics(self, y1, y2, metric):
        logger.debug('calculating comparison metrics')
        start_time = time.time()
        if len(y1) < 1 or len(y2) < 1:
            return {}

        ret = OrderedDict()

        min_l = min(len(y1), len(y2))
        #length of bot need be equal for AIC, BIC 
        y1 = y1#[:min_l]
        y2 = y2#[:min_l]

        #print('y1',len(y1))
        #print()
        #print('y2', len(y2))

        quantiles = np.linspace(start=0, stop=1, num=100)
        y11 = np.quantile(y1, quantiles, interpolation='linear')[25:76]
        y22 = np.quantile(y2, quantiles, interpolation='linear')[25:76]

        #logger.debug('before getting KS2samp test: ' + str(time.time() - start_time))
        try:
            ksi = scipy.stats.ks_2samp(y1, y2)
            ret['self-KS-2-samp-s'] = fmt(myconverter(ksi.statistic))
            ret['self-KS-2-samp-pvals'] = fmt(myconverter(ksi.pvalue))
        except:
            ret['self-KS-2-samp-s'] = '--'
            ret['self-KS-2-samp-pvals'] = '--'
        ##logger.debug('before getting rmse_of_quantiles betwee : ' + str(time.time() - start_time))
        #ret['nrmse_interquantiles'] =  str(self.get_rmse_of_quantiles(y1, y2))
        #logger.debug('before getting anderson-darling test: ' + str(time.time() - start_time))
        try:
            ret['anderson-darling test'] =  str(scipy.stats.anderson_ksamp([np.array(y1), np.array(y2)]))
        except:
            ret['anderson-darling test'] =  '--'
        ##logger.debug('before getting correlation coeff: ' + str(time.time() - start_time))
        # try:
        #     ret['correlation coeff'] = '--'
        # except:
        #     slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(y1[:min_l], y2[:min_l])
        #     ret['correlation coeff'] = str(r_value)
        ##logger.debug('before getting AIC: ' + str(time.time() - start_time))
        # try:
        #     ret['AIC'] = str(aic(y11, y22, 1)) #'skip? - very slow, also needs equal samples size' 
        # except:
        #     ret['AIC'] =  '--'
        ##logger.debug('before getting BIC: ' + str(time.time() - start_time))
        # try:
        #     ret['BIC'] = '--'#str(bic(y11, y22, 1)) #requires equal sample size
        # except:
        #     ret['BIC'] =  '--'
        return ret

    def get_comp_metrics_usess(self, usess_xys1, usess_xys2, metric):

        start_time = time.time()

        ret = OrderedDict()
        for m in ['self-KS-2-samp-s', 'self-KS-2-samp-pvals']:#,'nrmse_interquantiles', 'anderson-darling-sval','anderson-darling-pval','AIC', 'BIC']:
            ret[m] = []
        for us1 in usess_xys1:
            for us2 in usess_xys2:
                x1, y1 = us1
                x2, y2 = us2
                if len(y1) < 1 or len(y2) < 1:
                    continue
                min_l = min(len(y1), len(y2))
                #length of bot need be equal for AIC, BIC 
                y1 = y1#[:min_l]
                y2 = y2#[:min_l]
                #print('y1',len(y1))
                #print()
                #print('y2', len(y2))
                quantiles = np.linspace(start=0, stop=1, num=100)
                y11 = np.quantile(y1, quantiles, interpolation='linear')[25:76]
                y22 = np.quantile(y2, quantiles, interpolation='linear')[25:76]

                #print('before getting KS2samp test: ' + str(time.time() - start_time))
                try:
                    ksi = scipy.stats.ks_2samp(y1, y2)
                    ret['self-KS-2-samp-s'].append(fmt(myconverter(ksi.statistic)) )  
                    ret['self-KS-2-samp-pvals'].append(myconverter(fmt(ksi.pvalue))   )
                except :
                    #print('error')
                    ret['self-KS-2-samp-s'].append(None)   
                    ret['self-KS-2-samp-pvals'].append(None)   
                #print('before getting rmse_of_quantiles betwee : ' + str(time.time() - start_time))
                #ret['nrmse_interquantiles'].append(str(self.get_rmse_of_quantiles(y1, y2)))
                #print('before getting anderson-darling test: ' + str(time.time() - start_time))
                # try:
                #     ander = scipy.stats.anderson_ksamp([np.array(y1), np.array(y2)])
                #     ret['anderson-darling-sval'].append(ander.statistic)
                #     ret['anderson-darling-pval'].append(ander.pvalue)
                # except:
                #     ret['anderson-darling-sval'].append(None)
                #     ret['anderson-darling-pval'].append(None)
                # print('before getting correlation coeff: ' + str(time.time() - start_time))
                # try:
                #     ret['correlation coeff'] = '--'   
                # except:  
                #     slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(y1[:min_l], y2[:min_l])
                #     ret['correlation coeff'] = str(r_value)
                # print('before getting AIC: ' + str(time.time() - start_time))
                # try:
                #     ret['AIC'].append(aic(y11, y22, 1)) #'skip? - very slow, also needs equal samples size' 
                # except:
                #     ret['AIC'].append(None)
                # print('before getting BIC: ' + str(time.time() - start_time))
                # try:
                #     ret['BIC'].append(None)#bic(y11, y22, 1)) #requires equal sample size
                # except:
                #     ret['BIC'].append(None)
        return ret

    def get_rmse_of_quantiles(self, y1, y2):
        #normalized rmse of interquartile quantile differences
        quantiles = np.linspace(start=0, stop=1, num=100)

        y1_quantiles = np.quantile(y1, quantiles, interpolation='linear')
        y2_quantiles = np.quantile(y2, quantiles, interpolation='linear')

        amin = min(y1_quantiles[25], y2_quantiles[25])
        amax = max(y1_quantiles[76], y2_quantiles[76])

        if y1_quantiles[76] - y1_quantiles[25] ==0 or y2_quantiles[76] - y2_quantiles[25] ==0:
            return 'skipping'

        y1_int_quants = y1_quantiles[25:76]/(amax-amin)
        y2_int_quants = y2_quantiles[25:76]/(amax-amin)

        q = mean_squared_error(y_true=y1_int_quants, y_pred=y2_int_quants)
        return q
